#include <SPI.h>
#include <Adafruit_GFX.h>    // Core graphics library
#include <Adafruit_ST7735.h> // Hardware-specific library

#define TFT_CS 5
#define TFT_RST 7
#define TFT_DC 6

#define GAP_SIZE 40
#define HALF_GAP_SIZE 20
#define PIPE_NO 3
#define PIPE_WIDTH 10
#define PIPE_DISTANCE 25
#define BIRD_SIZE 5
#define ACCEPTABLE 60
#define WIN_SCORE 100

typedef struct pipe {
	uint8_t pos;
	uint8_t gap_pos;
} Pipe;

typedef struct bird { 
	uint8_t x;
	uint8_t y;
} Bird;

void game();
void print_text(byte x_pos, byte y_pos, char *text, byte text_size, uint16_t color);
void print_pipe(Pipe pipe, uint16_t color);
void get_pipes();
void print_pipes();
void print_bird(Bird bird, bool del);
bool check_collision(Bird bird);
void setup_game();
void print_score();
void init_jump_button();
uint8_t read_jump_button();
void init_buzzer(unsigned long freq);
void play_buzzer(unsigned long duration);
void play_buzzer();

Adafruit_ST7735 tft = Adafruit_ST7735(TFT_CS, TFT_DC, TFT_RST);

unsigned long lastDebounceTime = 0;
unsigned long debounceDelay = 5;
uint8_t buttonState; 
uint8_t lastButtonState = HIGH;
uint8_t fly_speed = 1;
uint8_t fall_speed = 2;
Pipe pipes[PIPE_NO];
Bird bird;
unsigned long last_time;
bool playing = true;
uint8_t score = 0;
uint8_t last_pipe_gap = 200;

void setup(void) {
	last_time = millis();
	randomSeed(analogRead(0));
	tft.initR(INITR_144GREENTAB); // Init ST7735R chip, green tab
	tft.setRotation(0); // set display orientation 
	Serial.begin(9600);
	init_jump_button();
	init_buzzer(1000);
	setup_game();
}

void setup_game() {
	score = 0;
	last_pipe_gap = 200;
	tft.fillScreen(ST77XX_BLACK);
	get_pipes();
	print_pipes();
	bird.x = 5;
	bird.y = 72;
	print_bird(bird, true);
	print_score();
}

void loop() {
	unsigned long crt = millis();
	if (crt - last_time >= 50 && playing) {
		game();
		last_time = crt;
	}

	if (!playing && read_jump_button() == LOW) {
		setup_game();
		playing = true;
	}
	
}

void game() {
	uint8_t reading = read_jump_button();

	if (reading != lastButtonState) {
		lastDebounceTime = millis();
	}

	bool jump = false;
	if ((millis() - lastDebounceTime) > debounceDelay) {
		if (reading != buttonState) {
			buttonState = reading;
			if (buttonState == LOW) {
				print_bird(bird, false);
				jump = true;
				play_buzzer(10);
			}
		}
	}
	
	lastButtonState = reading;
	print_bird(bird, false);
	bird.x += fly_speed;
	bird.y += fall_speed;
	if (jump) {
		bird.y -= 20;
	}
	if (bird.x >= 143) {
		bird.x = 1;
		tft.fillScreen(ST77XX_BLACK);
		get_pipes();
		print_pipes();
		print_score();
	}
	print_bird(bird, true);

	if (bird.x == 50 || bird.x == 85 || bird.x == 120) {
		score++;
		print_score();
	}


	if (check_collision(bird)) {
		delay(250);
		tft.fillScreen(ST77XX_BLACK);
		print_text(30, 40, "GAME", 3, ST77XX_RED);
		print_text(30, 70, "OVER", 3, ST77XX_RED);
		char score_text[4];
		sprintf(score_text, "%hhu", score);
		score_text[3] = '\0';
		print_text(55, 100, score_text, 3, ST77XX_CYAN);
		playing = false;
	}

	if (score == WIN_SCORE) {
		delay(250);
		tft.fillScreen(ST77XX_BLACK);
		print_text(30, 40, "YOU", 3, ST77XX_GREEN);
		print_text(30, 70, "WON!", 3, ST77XX_GREEN);
		char score_text[4];
		sprintf(score_text, "%hhu", score);
		score_text[3] = '\0';
		print_text(55, 100, score_text, 3, ST77XX_CYAN);
		playing = false;
	}
}

void print_text(byte x_pos, byte y_pos, char *text, byte text_size, uint16_t color) {
	tft.setCursor(x_pos, y_pos);
	tft.setTextSize(text_size);
	tft.setTextColor(color);
	tft.setTextWrap(true);
	tft.print(text);
}

void print_score() {
	char score_text[4];
	if (score != 0) {
		sprintf(score_text, "%hhu", score - 1);
		score_text[3] = '\0';
		print_text(0, 5, score_text, 2, ST77XX_BLACK);
	}
	memset(score_text, 0, sizeof(score_text));
	sprintf(score_text, "%hhu", score);
	score_text[3] = '\0';
	print_text(0, 5, score_text, 2, ST77XX_GREEN);
}

void get_pipes() {
	for (uint8_t i = 0; i < PIPE_NO; i++) {
		if (last_pipe_gap == 200) {
			pipes[i].pos = (PIPE_WIDTH + PIPE_DISTANCE) * (i + 1);
			pipes[i].gap_pos = (uint8_t) (HALF_GAP_SIZE + random(101));
			last_pipe_gap = pipes[i].gap_pos;
		} else {
			uint8_t ll = last_pipe_gap > ACCEPTABLE ? last_pipe_gap - ACCEPTABLE : HALF_GAP_SIZE;
			uint8_t hl = last_pipe_gap + ACCEPTABLE < 111 ? last_pipe_gap + ACCEPTABLE : 111;
			pipes[i].pos = (PIPE_WIDTH + PIPE_DISTANCE) * (i + 1);
			pipes[i].gap_pos = (uint8_t) (ll + random(hl - ll));
			last_pipe_gap = pipes[i].gap_pos;
		}
		
	}
}

void print_pipe(Pipe pipe, uint16_t color) {
	tft.fillRect(pipe.pos, 0, PIPE_WIDTH, pipe.gap_pos - HALF_GAP_SIZE, color);
	tft.fillRect(pipe.pos, pipe.gap_pos + GAP_SIZE / 2, PIPE_WIDTH, 144 - pipe.gap_pos - HALF_GAP_SIZE, color);
}

void print_pipes() {
	for (uint8_t i = 0; i < PIPE_NO; i++) {
		print_pipe(pipes[i], ST77XX_WHITE);
	}
}

void print_bird(Bird bird, bool paint) {
	if (paint) {
		tft.fillRect(bird.x - 2, bird.y - 2, BIRD_SIZE, BIRD_SIZE, ST77XX_BLUE);
	} else {
		tft.fillRect(bird.x - 2, bird.y - 2, BIRD_SIZE, BIRD_SIZE, ST77XX_BLACK);
	}
	
}

bool check_collision(Bird bird) {
	uint8_t ind = 0;
	if (bird.x < 50) {
		ind = 0;
	} else if (bird.x < 85) {
		ind = 1;
	} else {
		ind = 2;
	}

	uint8_t r2x = pipes[ind].pos;
	uint8_t r2y = pipes[ind].gap_pos - HALF_GAP_SIZE;
	uint8_t r2w = PIPE_WIDTH;
	uint8_t r2h = GAP_SIZE;

	uint8_t r1x = bird.x - 2;
	uint8_t r1y = bird.y - 2;
	uint8_t r1w = BIRD_SIZE;
	uint8_t r1h = BIRD_SIZE;
	if (r1x < r2x + r2w && r1x + r1w > r2x) {
		if (r1y >= r2y && r1y < r2y + r2h && r1y + r1h > r2y && r1y + r1h <= r2y + r2h) {
			return false;
		}
		return true;
	}
	return false;
}

void init_jump_button() {
	DDRD &= ~(1 << PD2);
    PORTD |= (1 << PD2);
}

uint8_t read_jump_button() {
	return PIND & (1 << PD2);
}

void init_buzzer(unsigned long freq) {
	DDRB |= (1 << PB1);

	TCCR1A = (1 << COM1A1) | (1 << WGM11);
    TCCR1B = (1 << WGM13) | (1 << WGM12) | (1 << CS10);
	ICR1 = F_CPU / freq - 1;

	TCCR1A &= ~(1 << COM1A1);
    PORTB &= ~(1 << PB1); 
	OCR1A = 0;
}

void play_buzzer(unsigned long duration) {
	unsigned long tm = millis();
	TCCR1A |= (1 << COM1A1);
	OCR1A = ICR1 / 2;
	while (millis() - tm < duration);
	TCCR1A &= ~(1 << COM1A1);
    PORTB &= ~(1 << PB1); 
	OCR1A = 0;
}